{-# OPTIONS_GHC -fno-warn-incomplete-patterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE FlexibleInstances #-}
module Lam.Render
  ( render
  ) where

import Lam.Adapters
import Lam.Binding
import Lam.Lam

import Data.Cursor.CLASE.Language
import Data.Cursor.CLASE.Traversal

import Control.Monad.Reader
import Control.Monad.Writer
import Data.Char
import Data.Map (Map)
import Data.Maybe
import Data.Sequence (Seq)
import qualified Data.Foldable as Seq
import qualified Data.Map as Map
import qualified Data.Sequence as Seq

newtype LRM a = LRM 
  { runLRM :: ReaderT (Map Integer String) (Writer (Seq Char)) a }
  deriving (MonadReader (Map Integer String), Monad, MonadWriter (Seq Char))

instance LamTraversalAdapterLam (LRM ()) where
  visitLam _ hole = hole

instance LamTraversalAdapterExp (LRM ()) where
  visitAbs (Abs name _ _) ty exp 
    = out ("λ " ++ name ++ "∷") >> ty >> out " . " >> exp
  visitApp _ l r 
    = out "(" >> l >> out "∘" >> r >> out ")"
  visitVar (Var i) 
    = (out . fromMaybe "Variable free!" =<< lookupBinding i) >> 
      (out . subscript $ i)
  visitNoExp _ = out "?"

instance LamTraversalAdapterType (LRM ()) where
  visitUnit _ = out "τ"
  visitArr _ lhs rhs = out "(" >> lhs >> out " → " >> rhs >> out ")"
  visitNoType _ = out "?"

instance LamTraversalAdapterCursor (LRM ()) where
  visitCursor _ child = out "⟦" >> child >> out "⟧"

subscript :: Integer -> String
subscript = map (chr . (+) (8320 - (ord '0')) . ord) . show 

render :: Cursor Lam x a -> String
render 
  = (\((),s) -> Seq.toList s) . runWriter 
  . flip runReaderT Map.empty . runLRM 
  . completeTraversal

out ::  [Char] -> LRM ()
out = tell . Seq.fromList
